/************************************************************************\
* Project Von Neumann - fe.c code file                                   *
* -----------------------------------------------------------------------*
* Copyright (C) 1995, 1996 Project Von Neumann Team.                     *
* -----------------------------------------------------------------------*
* Last Revised: May '96                                                  *
* By          : Martijn Faassen                                          *
* Version #   : 0.75                                                     *
* Description : Front end functions using Allegro or Linux svgalib or    *
*               even text based output!                                  *
\************************************************************************/

/* fever.h : get #defines for front end version */
#include "fever.h"

#ifdef ALLEGRO
#include "allegro.h" 
#include "demo.h" /* for data.dat test */
#endif

#ifdef SVGALIB
#include <math.h> /* for trig */
#include <vga.h>
#include <vgagl.h>
#include <vgakeyboard.h>
#endif

#ifdef TEXTOUT
#include <math.h>
#include <stdio.h>
#ifndef NOKEYS
#ifdef VGALIB_KEYS
#include <vgakeyboard.h>
#endif /* VGALIB_KEYS */
#ifdef ALLEGRO_KEYS
#include <allegro.h>
#endif /* ALLEGRO_KEYS */
#endif /* NOKEYS */
#endif /* TEXTOUT */ 

/* PVN include files */
#include "fe.h"       /* front end header */
#include "untypes.h"

/* array of keyboard flags. */
#ifdef ALLEGRO
extern volatile char key[128];  /* ALLEGRO */
#endif

#ifdef SVGALIB
char* key; /* will point to key buffer */
#endif

#ifdef TEXTOUT
#ifndef NOKEYS
#ifdef SVGALIB_KEYS
char* key;
#endif /* SVGALIB_KEYS */
#ifdef ALLEGRO_KEYS
#endif /* ALLEGRO_KEYS */
extern volatile char key[128];
#endif /* NOKEYS */
#endif /* TEXTOUT */

/* the screen bitmap */
#ifdef ALLEGRO
extern BITMAP* screen; /* pointer to the screen */
BITMAP* buffer; /* the screen buffer */
BITMAP* ship;  /* holds the ship bitmap */
#endif

#ifdef SVGALIB
GraphicsContext* screen; /* will point to the screen */
GraphicsContext screen_s;
unsigned char ship[32 * 32]; /* holds the ship bitmap */
unsigned char test[1 * 1];
#endif

#ifdef TEXTOUT
unsigned char ship[32 * 32]; /* dummy ship bitmap */
#endif

/* Trig lookup tables */
extern float UN_Sin[NUM_DIRS];
extern float UN_Cos[NUM_DIRS];

/* the dirty rectangle stack. */
static FE_DirtyRectStack DirtyStack[MAX_DISPLAY];
static unsigned int DirtyStackCounter;

/* the bitmap stack. */
static FE_BitmapStack BitmapStack[MAX_DISPLAY];
static unsigned int BitmapStackCounter;

/* the rescaling lookup table for 1024x768 -> 800x600 */
/* static unsigned short ReScale800x600[1024]; */
/* the rescaling lookup table for 1024x768 -> 640x480 */
static unsigned short ReScale640x480[1024];

/**************************************************************************\
* Function Name: FE_Init
* Description  : Initializes the front end.
* Arguments    : none
* Returns      : void
\**************************************************************************/
void FE_Init()
     {
#ifdef ALLEGRO
	/* initializes the allegro library. This must be called before
	 * anything else allegro related is called. */
	allegro_init(); /* ALLEGRO */
	/* installs the keyboard handler */
	install_keyboard(); /* ALLEGRO */
	/* sets the gfx mode */
	set_gfx_mode(GFX_S3, 640, 480, 0, 0);
	/* create buffer of same size as the screen */
	buffer = create_bitmap(640,480);
	clear(buffer);
	/* temporary: creates a ship bitmap */
	 
	ship = create_bitmap(32, 32);
	clear(ship);
	vline(ship, 15, 0, 31, 245);         
	hline(ship, 7, 30, 23, 244);
	
#endif /* ALLEGRO */
	
#ifdef SVGALIB
	int i;
	/* first do graphics stuff. In case of wrong I/O permissions
	 * it will hopefully bail out before the evil keyboard driver
	 * is installed. */
	if (vga_hasmode(G640x480x256))
	  vga_setmode(G640x480x256);
	else
	  {
	     FE_Exit();
	  }
	
	/* keyboard stuff */
#ifndef NOKEYS
	keyboard_init(); /* initialize keyboard */
	keyboard_setdefaulteventhandler(); /* set default eventhandler */
	keyboard_translatekeys(TRANSLATE_CURSORKEYS); /* map cursor keys to numeric keypad cursor */
	keyboard_clearstate();  /* clear keyboard buffer */
	key = keyboard_getstate(); /* let key point to keyboard buffer */
#endif /* NOKEYS */
	
	/* set context to screen */
	gl_setcontextvga(G640x480x256);
	/* screen points at a structure */
	screen = &screen_s;
	
	/* pointer the current graphics context in 'screen' */
	gl_getcontext(screen);
	
	/* make virtual screen the same size as actual screen */
	gl_setcontextvgavirtual(G640x480x256);
	
	gl_enableclipping(); /* clipping on */
	gl_setclippingwindow(0, 0, 639, 479); 
	
	/* make test bitmap */
	for (i = 0; i < (32 * 32); i++)
	  {
	     ship[i] = 34; /* set ship to block of single color */
	     
	  }
	
	test[0] = 34;
	
	for (i = 0; i < MAX_DISPLAY; i++)
	  {
	     BitmapStack[i].x = 0;
	     BitmapStack[i].y = 0;
	     BitmapStack[i].w = 0;
	     BitmapStack[i].h = 0;
	     BitmapStack[i].r = 0;
	     BitmapStack[i].bitmap = 0;
	     
	     DirtyStack[i].x = 0;
	     DirtyStack[i].y = 0;
	     DirtyStack[i].w = 0;
	     DirtyStack[i].h = 0;
	  }
	
#endif /* SVGALIB */
	
#ifdef TEXTOUT

#ifndef NOKEYS
#ifdef SVGALIB_KEYS
	keyboard_init(); /* initialize keyboard */
	keyboard_setdefaulteventhandler(); /* set default eventhandler */
	keyboard_translatekeys(TRANSLATE_CURSORKEYS); /* map cursor keys to numeric keypad cursor */
	keyboard_clearstate();  /* clear keyboard buffer */
	key = keyboard_getstate(); /* let key point to keyboard buffer */
#endif /* SVGALIB_KEYS */
#ifdef ALLEGRO_KEYS
	install_keyboard(); /* initialize keyboard */
#endif /* ALLEGRO_KEYS */
#endif /* NOKEYS */

	printf("Front end is now in Text Output mode.\n");
	
#endif /* TEXTOUT */
	
	DirtyStackCounter = 0;
	BitmapStackCounter = 0;
	
	FE_InitTrig();
	FE_InitReScale();
     }

/**************************************************************************\
* Function Name: FE_Exit
* Description  : Shuts down the FE, and ends the program.
* Arguments    : none
* Returns      : void
* Notes        : this function must be called last.
\**************************************************************************/
void FE_Exit()
     {
#ifdef ALLEGRO
	/* removes buffer bitmap from memory */
	destroy_bitmap(buffer); /* ALLEGRO */
	/* this removes any keyboard, mouse, and timer routines allegro
	 * installed previously. */
	allegro_exit(); /* ALLEGRO */
#endif
	
#ifdef SVGALIB
#ifndef NOKEYS
	keyboard_close(); /* keyboard back to normal */
#endif /* NOKEYS */
	vga_setmode(TEXT); /* back to text mode */
	gl_setcontextvga(TEXT);
#endif /* SVGALIB */
	
#ifdef TEXTOUT
#ifndef NOKEYS
#ifdef SVGALIB_KEYS
	keyboard_close();
#endif /* SVGALIB_KEYS */
#ifdef ALLEGRO_KEYS
	remove_keyboard();
#endif /* ALLEGRO_KEYS */
#endif /* NOKEYS */
#endif /* TEXTOUT */
	
	exit(0);  /* the end */
     }

/***************************************************************************\
* Function Name: FE_InitTrig
* Description  : Initializes UN_Sin and UN_Cos arrays.
* Arguments    : none
* Returns      : void
\***************************************************************************/
void FE_InitTrig()
     {
	int i;
	
#ifdef ALLEGRO
	for (i = 0; i < NUM_DIRS; i++)
	  {
	     UN_Sin[i] = fixtof(fsin(itofix(i))); 
	     UN_Cos[i] = fixtof(fcos(itofix(i)));
	  }
#endif /* ALLEGRO */
	
#ifdef SVGALIB
	for (i = 0; i < NUM_DIRS; i++)
	  {
	     UN_Sin[i] = (float) sin((double)i * 
				     ((double)TWOPI / (double)NUM_DIRS));
	     UN_Cos[i] = (float) cos((double)i * 
				     ((double)TWOPI / (double)NUM_DIRS));
	  }
#endif /* SVGALIB */
	
#ifdef TEXTOUT
	for (i = 0; i < NUM_DIRS; i++)
	  {
	     UN_Sin[i] = (float) sin((double)i * 
				     ((double)TWOPI / (double)NUM_DIRS));
	     UN_Cos[i] = (float) cos((double)i * 
				     ((double)TWOPI / (double)NUM_DIRS));
	  }
#endif /* TEXTOUT */
     }

/***************************************************************************\
* Function Name: FE_InitReScale
* Description  : Initializes rescaling lookup tables.
* Arguments    : none
* Returns      : void
* Notes        : To limit memory we want to adapt this function so it will
*                only initialize the necessary arrays.
*                This is based on the display resolution.
\***************************************************************************/
void FE_InitReScale()
     {
	unsigned int i;
	
	/*
	for (i = 0; i < 1024; i++)
	  ReScale800x600[i] = (unsigned short)((float)i/(float)1.28);
	*/
	
	for (i = 0; i < 1024; i++)
	  ReScale640x480[i] = (unsigned short)((float)i/(float)1.6); 
     }


/***************************************************************************\
* Function Name: FE_GetActions
* Description  : Checks the actions the player issues at this moment and
*                returns an actions bitstring based on that.
* Arguments    : none
* Returns      : UN_Actions
* Notes        : this currently only checks keys pressed, not the joystick
*                or mouse.
*                Presumably a function AL_GetActions(UN_Object object)
*                exists to get non player animate actions.
*                This needs a fix so as to return either left or right,
*                not both at once.
\***************************************************************************/
UN_Actions FE_GetActions()
     {
	UN_Actions pressed = 0;
	
#ifdef ALLEGRO
	pressed |= (key[KEY_UP]    ? ACT_THRUST : 0);
	pressed |= (key[KEY_DOWN]  ? ACT_BRAKE  : 0);
	pressed |= (key[KEY_LEFT]  ? ACT_LEFT   : 0);
	pressed |= (key[KEY_RIGHT] ? ACT_RIGHT  : 0);
#endif
	
	
#ifdef SVGALIB

#ifndef NOKEYS
	keyboard_update(); /* update keyboard */
	
	pressed |= (key[SCANCODE_CURSORUP]    ? ACT_THRUST : 0);
	pressed |= (key[SCANCODE_CURSORDOWN]  ? ACT_BRAKE  : 0);
	pressed |= (key[SCANCODE_CURSORLEFT]  ? ACT_LEFT   : 0);
	pressed |= (key[SCANCODE_CURSORRIGHT] ? ACT_RIGHT  : 0);
#endif /* NOKEYS */
	
#endif /* SVGALIB */
	
#ifdef TEXTOUT

#ifndef NOKEYS
#ifdef SVGALIB_KEYS
	keyboard_update(); /* update keyboard */
	
	if (key[SCANCODE_CURSORUP])
	  printf("Pressed up!\n");
	if (key[SCANCODE_CURSORDOWN])
	  printf("Pressed down!\n");
	if (key[SCANCODE_CURSORLEFT])
	  printf("Pressed left!\n");
	if (key[SCANCODE_CURSORRIGHT])
	  printf("Pressed right!\n");
	
	pressed |= (key[SCANCODE_CURSORUP]    ? ACT_THRUST : 0);
	pressed |= (key[SCANCODE_CURSORDOWN]  ? ACT_BRAKE  : 0);
	pressed |= (key[SCANCODE_CURSORLEFT]  ? ACT_LEFT   : 0);
	pressed |= (key[SCANCODE_CURSORRIGHT] ? ACT_RIGHT  : 0);

	printf("Pressed: %lu\n", pressed);
#endif /* SVGALIB_KEYS */     
#ifdef ALLEGRO_KEYS
   
	if (key[KEY_UP])
	  printf("Pressed up!\n");
	if (key[KEY_DOWN])
	  printf("Pressed down!\n");
	if (key[KEY_LEFT])
	  printf("Pressed left!\n");
	if (key[KEY_RIGHT])
	  printf("Pressed right!\n");

	pressed |= (key[KEY_UP]    ? ACT_THRUST : 0);
	pressed |= (key[KEY_DOWN]  ? ACT_BRAKE  : 0);
	pressed |= (key[KEY_LEFT]  ? ACT_LEFT   : 0);
	pressed |= (key[KEY_RIGHT] ? ACT_RIGHT  : 0);

	printf("Pressed: %lu\n", pressed);
#endif /* ALLEGRO_KEYS */
#endif /* NOKEYS */
	
	
#endif /* TEXTOUT */
	
       return pressed;
     }

/************************************************************************\
* Function Name: FE_PushDirtyRect
* Description  : Pushes a rectangular area on the display to the dirty
*                rectangle stack.
* Arguments    : x, y - top left coordinates of rectangle
*                w, h - width and height of rectangle
* Returns      : void
* Notes        : Utter disaster happens if DirtyStackCounter >= MAX_DISPLAY
\************************************************************************/
void FE_PushDirtyRect(int x, int y, int w, int h)
     {
	DirtyStackCounter++; /* update stack counter */
	
	DirtyStack[DirtyStackCounter].x = x;
	DirtyStack[DirtyStackCounter].y = y;
	DirtyStack[DirtyStackCounter].w = w;
	DirtyStack[DirtyStackCounter].h = h;
#ifdef TEXTOUT
	printf("PushDirtyRect: x:%d y:%d w:%d h:%d\n", x, y, w, h);
#endif /* TEXTOUT */
     }

/*************************************************************************\
* Function Name: FE_PushBitmap
* Description  : Pushes a bitmap to draw on the bitmap stack.
* Arguments    : x, y - where to draw (top left corner of bitmap)
*                w, h - width and height of bitmap
*                r    - possible rotation (0..255) 0 means no rotation
*                bitmap - pointer to the bitmap structure to draw
* Returns      : void
* Notes        : Rescales incoming 1024x768 coordinates to
*                physical screen coordinates.
\*************************************************************************/
void FE_PushBitmap(int x, int y, int w, int h,
		   unsigned char r, void* bitmap)
     {
	BitmapStackCounter++; /* update stack counter */
	
	BitmapStack[BitmapStackCounter].x = ReScale640x480[x];
	BitmapStack[BitmapStackCounter].y = ReScale640x480[y];
	BitmapStack[BitmapStackCounter].w = w;
	BitmapStack[BitmapStackCounter].h = h;
	BitmapStack[BitmapStackCounter].r = r;
#ifndef TEXTOUT
	BitmapStack[BitmapStackCounter].bitmap = bitmap;
#endif
#ifdef TEXTOUT
	printf("PushBitmap: x:%d y:%d w:%d h:%d r:%d\n", x, y, w, h, r);
#endif
     }

/*************************************************************************\
* Function Name: FE_ClearAllDirtyRect
* Description  : This function clears all dirty rectangles that are on
*                the stack.
* Arguments    : none
* Returns      : void
* Notes        : DirtyStackCounter is automatically set to 0 by this.
\*************************************************************************/
void FE_ClearAllDirtyRect()
     {
	for (;DirtyStackCounter != 0; DirtyStackCounter--)
	  {
	     
	     /* fill rectangular area on buffer with color 0 */
#ifdef ALLEGRO
	     rectfill(buffer, 
		      DirtyStack[DirtyStackCounter].x, 
		      DirtyStack[DirtyStackCounter].y,
		      DirtyStack[DirtyStackCounter].x + DirtyStack[DirtyStackCounter].w,
		      DirtyStack[DirtyStackCounter].y + DirtyStack[DirtyStackCounter].h,
		      0); 
#endif /* ALLEGRO */
	     
#ifdef SVGALIB
	     
	     gl_fillbox(DirtyStack[DirtyStackCounter].x,
			DirtyStack[DirtyStackCounter].y,
			DirtyStack[DirtyStackCounter].w,
			DirtyStack[DirtyStackCounter].h,
			0);
	      
#endif /* SVGALIB */

#ifdef TEXTOUT
	     printf("Clear dirty rect: x:%d y:%d w:%d h:%d\n",
		    DirtyStack[DirtyStackCounter].x,
		    DirtyStack[DirtyStackCounter].y,
		    DirtyStack[DirtyStackCounter].w,
		    DirtyStack[DirtyStackCounter].h);
#endif /* TEXTOUT */
	  }
     }

/***************************************************************************\
* Function Name: FE_DrawAllBitmap
* Description  : Draws all pictures as stored in the DrawStack. Rotates
*                if indicated by DrawStack.
* Arguments    : none
* Returns      : void
* Notes        : Rotation is currently not implemented for the
*                SVGALIB version.
\***************************************************************************/
void FE_DrawAllBitmap()
     {
	for (;BitmapStackCounter != 0; BitmapStackCounter--)
	  {
	     /* put what we are going to draw on the DirtyStack */
	     FE_PushDirtyRect(BitmapStack[BitmapStackCounter].x,
			      BitmapStack[BitmapStackCounter].y,
			      BitmapStack[BitmapStackCounter].w,
			      BitmapStack[BitmapStackCounter].h);
	     
#ifdef ALLEGRO
	     #ifdef 0
	     if (BitmapStack[BitmapStackCounter].r == 0)
	       {
		  /* we do not have to rotate */
		  draw_sprite(buffer,
			      BitmapStack[BitmapStackCounter].bitmap,
			      BitmapStack[BitmapStackCounter].x,
			      BitmapStack[BitmapStackCounter].y);
	       }
	     else
	       {
	       #endif 
		  /* draw rotated */
		  rotate_sprite(buffer,
			       BitmapStack[BitmapStackCounter].bitmap,
			       BitmapStack[BitmapStackCounter].x,
			       BitmapStack[BitmapStackCounter].y,
			       itofix(BitmapStack[BitmapStackCounter].r));
	      /* }  */                
#endif /* ALLEGRO */

#ifdef SVGALIB
	     gl_putbox(BitmapStack[BitmapStackCounter].x,
		       BitmapStack[BitmapStackCounter].y,
		       BitmapStack[BitmapStackCounter].w,
		       BitmapStack[BitmapStackCounter].h,
		       BitmapStack[BitmapStackCounter].bitmap);

#endif /* SVGALIB */

#ifdef TEXTOUT
	     printf("Draw bitmap: x:%d y:%d w:%d h:%d r:%d\n",
		    BitmapStack[DirtyStackCounter].x,
		    BitmapStack[DirtyStackCounter].y,
		    BitmapStack[DirtyStackCounter].w,
		    BitmapStack[DirtyStackCounter].h,
		    BitmapStack[DirtyStackCounter].r);
#endif /* TEXTOUT */
	  }
     }

/**************************************************************************\
* Function Name: FE_UpdateScreen
* Description  : - Cleans the buffer
*                - Updates the buffer
*                - blits the buffer to screen
* Arguments    : none
* Returns      : void
\**************************************************************************/
void FE_UpdateScreen()
     {
	/* clean the buffer by removing all dirty rectangles */
	FE_ClearAllDirtyRect();
	/* dirty stack is now empty */
	
	/* draw all bitmaps on the bitmap stack */
	FE_DrawAllBitmap();
	/* dirty stack now contains all dirty rectangles again */
	
	/* now update the physical screen with the buffer */
	
#ifdef ALLEGRO
	/* copy the buffer to the screen */
	blit(buffer, screen, 0, 0, 0, 0, 640, 480); 
#endif /* ALLEGRO */
	
#ifdef SVGALIB
	/* copy current graphics context to screen */
	gl_copyscreen(screen); 
#endif /* SVGALIB */
	
#ifdef TEXTOUT
	printf("Updating Screen..\n");
#endif /* TEXTOUT */
     }


/***************************************************************************\
* Function name: FE_GetEsc
* Description  : Check if the escape key is pressed. If it is, then 
*                the FE is shut down.
* Arguments    : none
* Returns      : void
* Notes        : Temporary function.
\***************************************************************************/
void FE_GetEsc()
     {

#ifdef ALLEGRO
	if (!key[KEY_ESC])
	  return;
	FE_Exit();
#endif /* ALLEGRO */
	
#ifdef SVGALIB
#ifndef NOKEYS
	if (!key[SCANCODE_ESCAPE])
	  return;
	FE_Exit();
#endif /* NOKEYS */
#endif /* SVGALIB */ 

#ifdef TEXTOUT
#ifndef NOKEYS
#ifdef SVGALIB_KEYS
	if (!key[SCANCODE_ESCAPE])
	  return;
	FE_Exit();
#endif /* SVGALIB_KEYS */
#ifdef ALLEGRO_KEYS
	if (!key[KEY_ESC])
	  return;
	FE_Exit();
#endif /* ALLEGRO_KEYS */
#endif /* NOKEYS */
#endif /* TEXTOUT */
     }
	
