#ifndef FE_H
#define FE_H
/************************************************************************\
* Project Von Neumann - fe.h header file                               *
* -----------------------------------------------------------------------*
* Copyright (C) 1995, 1996 Project Von Neumann Team.                     *
* -----------------------------------------------------------------------*
* Last Revised: May '96                                                  *
* By          : Martijn Faassen                                          *
* Version #   : 0.75                                                     *
* Description : Header file for front end                                *
\************************************************************************/ 

#include "untypes.h"

/**************************************************************************\
* Description: Defines used by FE
\**************************************************************************/

/* size of bitmap and dirty rectangle stacks. This is the maximum amount of
 * bitmaps the FE can display at the same time */
#define MAX_DISPLAY 500

/**************************************************************************\
* Description: Structures used by FE
\**************************************************************************/

/* for dirty rectangle stack */
typedef struct FE_DirtyRectStack_s
     {
        int x, y;
	int w, h;
     } FE_DirtyRectStack;

/* for bitmap stack */
typedef struct FE_BitmapStack_s
     {
	int x, y;
	int w, h;
	int r;
	void* bitmap;
     } FE_BitmapStack;

/****************************************************************************\
* Description: Front end function declarations.
\****************************************************************************/

void FE_Init(); /* initialize the FE */
void FE_Exit(); /* shutdown the FE */
void FE_InitTrig(); /* initialize UN_Cos and UN_Sin */
void FE_InitReScale(); /* initialize rescaling lookup table */

UN_Actions FE_GetActions();
void FE_PushDirtyRect(int x, int y, int w, int h);
void FE_PushBitmap(int x, int y, int w, int h, 
		   unsigned char r, void* bitmap);
void FE_ClearAllDirtyRect();
void FE_DrawAllBitmap();

void FE_UpdateScreen();
void FE_GetEsc();

#endif /* FE_H */
