#ifndef UNTYPES_H
#define UNTYPES_H
/************************************************************************\
* Project Von Neumann - untypes.h header file                            * 
* -----------------------------------------------------------------------* 
* Copyright (C) 1995, 1996 Project Von Neumann Team.                     *
* -----------------------------------------------------------------------*
* Last Revised: May '96                                                  *
* By          : Martijn Faassen                                          *
* Version #   : 0.75                                                     *
* Description : General stuff used by universe fuctions.                 *
\************************************************************************/ 

/************************************************************************\
* Description: Universe #defines
\************************************************************************/

/* TRUE and FALSE need to be defined if they are not defined already */
#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif 

/* The action keys */
#define ACT_THRUST       0x00000001
#define ACT_BRAKE        0x00000002
#define ACT_LEFT         0x00000004
#define ACT_RIGHT        0x00000008
#define ACT_NEXT_TARGET  0x00000010
#define ACT_PREV_TARGET  0x00000020
#define ACT_FIRE         0x00000040
#define ACT_EAT          0x00000080
#define ACT_NEXT_WEAPON  0x00000100
#define ACT_PREV_WEAPON  0x00000200
#define ACT_BREED        0x00000400

/* For ease of programming */
#define ANIMATE (object->extra.animate)

#define NUM_DIRS 256 /* amount of facing directions */

#define TWOPI 6.2831853

/* Max thrust. Currently it ranges from 0 to 100 
 * Note that we may change this to a parameter eventually. */
#define MAX_THRUST 100

/**************************************************************************\
* Description: Various typedefs
\**************************************************************************/

typedef unsigned long UN_Actions;   /* 32 bit number which stores actions */
typedef signed long UN_Coord;       /* universe coordinates max about 2^31 */
typedef float UN_Velocity;          /* object velocity */
typedef unsigned long UN_Direction; /* object direction */
typedef unsigned char UN_ObjType;   /* object type */
typedef unsigned long UN_Sector;    /* sector of playing field */
typedef unsigned char UN_Bool;      /* bool */

/***************************************************************************\
* Description: Enumeration of the object types.
\***************************************************************************/
enum  {
   OBJ_PLAYER = 0,
   OBJ_PROBE,
   OBJ_SUPPLY,
   OBJ_MISSILE, /* note that everything <= MISSILE is animate */
   OBJ_BULLET,
   OBJ_ASTEROID,
   OBJ_UF_EGG,
   OBJ_F_EGG,
   OBJ_D_EGG,
   OBJ_D_PLAYER,
   OBJ_D_PROBE,
   OBJ_D_SUPPLY,
   OBJ_UNMOVABLE,
   OBJ_STICKY,
   OBJ_DUMMY, 
   OBJ_L_OTYPE
};

/**************************************************************************\
* Description: Structures used by universe
\**************************************************************************/

/* Extra information for animate objects */
typedef struct UN_AnimateInfo_s
     {
	unsigned char turn;
	unsigned long thrust;
	UN_Actions actions;
     } UN_AnimateInfo;

/* Extra information for bullet objects */
typedef struct UN_BulletInfo_s
     {
     } UN_BulletInfo;

/* A universe object */
typedef struct UN_Object_s
     {
	UN_ObjType what; /* object type */
	UN_Coord x, y;  /* coordinates in universe */
	UN_Sector sectx, secty; /* sector we are in */
	UN_Velocity velx, vely; /* velocity vector */
	UN_Direction facedir; /* facing direction 0..255 */
	unsigned int w, h; /* object weight and height */
	long shield; /* shield strength */
	long energy; /* energy/nutricious value/explosive value */
	long mass;   /* mass of object */
        unsigned long age; /* age */

	/* this object is in a sector list */
	struct UN_Object_s* next_sect;
	struct UN_Object_s* prev_sect;
	/* this object is in the global active object list */
	struct UN_Object_s* next_glob;
	struct UN_Object_s* prev_glob;
	
	void* pict; /* pointer to the current picture */
	
	/* extra information, specific to object type */
	union
	  {
	     UN_AnimateInfo animate;
	     UN_BulletInfo bullet;
	  } extra;
	
     } UN_Object;

#endif
